function iroots(varargin)
% This program provides a user-friendly graphical frontend for Matlab and 
% Octave native 'fzero' to find roots of a function.  The program plots the 
% function to present the overview of all possible roots over an interval.
%
% Usage:
% 
%   iroots('t') or iroots('T')
%       Using templates
%
%   iroots(a, b)
%       Finding a root between the interval of [a, b], and plot the 
%       function with 10,000 points.  
%       If the orders of magnitude of a and b differ by 5 or larger, 
%       data points for the plot are generated evenly between log10(a) 
%       and log10(b), and the scale of x-axis is logarithmic.
%
%   iroots(a, b, nPoints)
%       Finding a root between the interval of [a, b], and plot the 
%       function with nPoints points.
%
%   iroots(x)
%       Finding a root around x.  x is an estimation of the root.
%
%   iroots
%       Finding a root between the interval of [1e-14, 1], and plot
%       the function with 10,000 points.
%       This is particularly useful to find concentrations of chemical 
%       species in a solution, since most concentrations are in such an 
%       interval.
%
% Remarks
%   1. Write your function in the m-file named "myfunction.m".
%
%   2. You should either zoom in or pan the plot to determine an interval
%      in which the curve in black traverses only once the horizontal
%      zero-line in red.  Meanwhile, the color of the axis turns green.
%
%   3. When you are satisfied with the interval, close the figure to
%      start the iteration.
%
%   4. If the condition in Remark #2 is not satisfied, the color of the
%      axis turns pink.  In this case, the action of figure closing just 
%      closes the figure without starting the iteration.
%
%   5. Due to the nature of digital computation, different runs of the 
%      same function could not yield the exact same results.
%
%   6. For a found root, the program evaluates the function at it.  
%      Based on the closeness of the evaluation to zero you can 
%      determine the accuracy of the root.
%

% The twentieth modification was on Tuesday of Tues. 26, 2024.
%   - handled the issue of low accuracy of roots smaller than 1e-14.
% The nineteenth modification was on Friday of Sept. 10, 2021.
%   - fixed some bugs.
% The eighteenth modification was on Thursday of Aug. 5, 2021.
%   - fixed some bugs.
% The seventeenth modification was on Friday of May 28, 2021.
%   - added the support of templates for common functions.
%   - clear global variables after the program ends.
% The sixteenth modification was on Thursday of Aug. 27, 2020.
%   - the program runs on Octave too, so it is open source.
% The fifteenth modification was on Monday of Oct. 21, 2019.
%   - fixed an error on the MacOS.
% The fourteenth modification was on Sunday of Sept. 1, 2019.
%   - avoid too long loops in function 'bingo'.
% The thirteenth modification was on Tuesday of Oct. 16, 2018.
%   - fixed the handling of roots smaller than double precision.
%   - added displaying the logarithm of the positive root. 
%   - fixed the enlargement of the interval with negative borders found 
%     by 'bingo'.
% The twelfth modification was on Tuesday of Jan. 2, 2018.
%   - update check was improved.
%   - better handling of roots smaller than double precision.
%   - data points are marked when zooming in renders fewer than 10 points
%     in the axes.
% The eleventh modification was on Tuesday of Sept. 6, 2016.
%   - set pre- and post-functions of 'zoom' and 'pan' in a compatible way 
%     with higher versions than R2010a.
%   - notify the update on exit even though the bisection does not start.
% The tenth modification was on Thursday of Jun. 2, 2016.
%   - added update check.
% The ninth modification was on Friday of May 6, 2016.
% The eighth modification was on Friday of Apr. 30, 2016. 
%   - added module "bingo".
% The seventh modification was on Wednesday of Apr. 2, 2016.
% The sixth modification was on Saturday of Feb. 27, 2016.
% The fifth modification was on Monday of Nov. 30, 2015.
% The fourth modificatoin was on Monday of Jun. 15, 2015.
% The third modification was on Tuesday of Feb. 17, 2015.
% The second modification was on Saturday of Jun. 7, 2014.
% The first modification was on Monday of Mar. 24, 2014.
% By Limin Shao on Wednesday of Mar. 5, 2014.
%
% Copyright 2021 Limin Shao.

global PermissionColor ProhibitionColor isNewVersionAvailable Webpage isOctave
PermissionColor = [0.8 1 0.8];
ProhibitionColor = [1 0.8 0.8];
isNewVersionAvailable = false;
Webpage = 'http://staff.ustc.edu.cn/~lshao/misc.html';
isOctave = exist('OCTAVE_VERSION', 'builtin') == 5;

WindowPositionScale = 0.8;

VersionControl = 'Finding roots of a function (v2024.r1)';

LowestPlatformVersion = (1 - isOctave) * 7.1 + isOctave * 5.2;
% the lowest version to run the program is 7.1 for Matlab, and 5.2 for
% Octave.
PlatformDigitalVersion = sscanf(version, '%f');
if PlatformDigitalVersion(1) < LowestPlatformVersion
    disp(' ');
    disp('Error!');
    disp(' ');
    if isOctave
        disp('The software only runs under Octave 5.2.0 or newer.');
    else
        disp('The software only runs under Matlab 7.10 (R2010a) or newer.');
    end
    disp(' ');
    input('Press any key to quit...  ');
    return;
end
% Check the Matlab main version.

% <<<< dealing with templates.
if (nargin == 1) && ((varargin{1} == 't') || (varargin{1} == 'T'))
    Templates = ProcessTemplates('iroots_templates.txt');
    if isempty(Templates)
        return;
    else
        nTemplates = length(Templates.Positions);
        disp(' ');
        disp([num2str(nTemplates) ' templates are available as']);  
        for i = 1: nTemplates
            disp(['  ' num2str(i) ': ' ...
                Templates.Contents{Templates.Positions(i), 1}(2:end)]);
        end
        
        disp(' ');
        Sequence = [];
        while isempty(Sequence)
            Sequence = input('which template do you want to load? ', 's');
        end
        % avoid accidental Enter without inputting anything.
        
        Sequence = str2double(Sequence);
        if isnan(Sequence) || (Sequence < 1) || (Sequence > nTemplates)
            disp(' ');
            disp('Invalid input.  The program stops.');
            return;
            % invalid input
        else
            TemplateStart = Templates.Positions(Sequence) + 1;
            if Sequence == nTemplates
                TemplateEnd = length(Templates.Contents);
            else
                TemplateEnd = Templates.Positions(Sequence + 1) - 1;
            end

            Target = ['%% Template: ' ...
                Templates.Contents{Templates.Positions(Sequence)}(2:end)];
            for i = TemplateStart:TemplateEnd
                Target = [Target '\n' Templates.Contents{i, 1}];
            end

            if isOctave
                disp(' ');
                disp('------------------------------------------');
                fprintf([Target '\n']);
                disp('------------------------------------------');
                disp(' ');
                disp('Copy and paste the content between the dash lines to "myfunction.m".');
            else
                clipboard('copy', sprintf(Target));
                disp(' ');
                disp(['Template #' num2str(Sequence) ' is copied to the clipboard.']);
                disp('Paste it to "myfunction.m".');
            end
            
            return;
            % the desired template is properly dealt.
        end
    end
end

% >>>> dealing with templates.


% <<<< check for update
aux = strfind(VersionControl, '(v') + 2;
CurrentVersion = VersionControl(aux:end-1);
aux = strfind(CurrentVersion, '.');
CurrentVersion(aux) = [];
% information for update check.

try
    webcontent = urlread(Webpage);
    aux1 = regexpi(webcontent, 'iroots_+[a-z, 0-9]+.zip', 'match');
    if isempty(aux1)
        FileName = [];
       % the web server mulfunctioned or the webpage was incorrect.
    else
        FileName = aux1{1};
    end
    
    if ~strcmp(FileName, ['iroots_' CurrentVersion '.zip'])
        isNewVersionAvailable = true;
    end
catch ME
    isNewVersionAvailable = false;
    % when the connection is off, new version cannot be determined.
end
% >>>> Check for update.

PS = '\';
if ismac
    PS = '/';
end
% set path separater for Windows and MacOS.
AbsolutePath = [pwd PS 'myfunction.m'];
if ~exist(AbsolutePath, 'file')
    disp(' ');
    disp('Error!  File ''myfunction.m'' is missing.');
    disp(' ');
    FileID = fopen(AbsolutePath, 'w');
    if FileID == -1
        disp('Unable to create a new ''myfuncion.m''.  Try following steps to solve the problem.');
        disp('Exit Matlab or Octave.  Right click on the shortcut or the executive file of ');
        disp('Matlab or Octave, then click item ''Run as administrator'' from the menu.');
        return;
    else
        fprintf(FileID, '%s\n\n', 'function y = myfunction(x)');
        fprintf(FileID, '%s\n', '% This is an example to calculate the [H+] of a 0.1 M HAc solution.');
        fprintf(FileID, '%s\n', 'y = x - 1.8e-5 * 0.1 / (x + 1.8e-5) - 1e-14 / x;');
        fclose(FileID);
        disp('A new ''myfunction.m'' is created for you to input the function.');
        edit myfunction
    end
    
    return;
end
% Check the existence of 'myfunction.m', and create one if it is absent.

nPoints = 10000;

switch nargin
    case 0
        a = 1e-14; b = 1;
    case 1
        Interval = bingo(varargin{1});
        if isempty(Interval)
            disp(' ');
            disp(['Error!  Could not find a valid interval for the bisection method around ' ...
                num2str(varargin{1}) '.']);
            disp(' ');
            return;
        else
            a = Interval(1); b = Interval(2);
        end
    case 2
        [a, b] = deal(varargin{:});
    case 3
        [a, b, nPoints] = deal(varargin{:});
    otherwise
        disp('Incorrect syntax.  Type ''help iroots'' for information.');
        return;
end
% Assign value to necessary parameters.

isLogarithmicX = false;
x = linspace(min(a, b), max(a, b), nPoints);

if (a > 0) && (b > 0)
    Exponent1 = log10(min(a, b)); Exponent2 = log10(max(a, b));
    if (Exponent2 - Exponent1) > 5
        x = 10 .^ linspace(Exponent1, Exponent2, nPoints);
        isLogarithmicX = true;
    end
end

y = zeros(size(x));

try
    for i = 1:nPoints
        y(i) = myfunction(x(i));
    end
catch ME
    disp(' ');
    disp('An error occurred when executing ''myfunction.m''.  The error message is');
    disp(' ');
    disp([' ' ME.message]);
    disp(' ');
    input('Press any key to quit...  ');
    return;
end
% This loop is time-consuming and awkward, but it relieves users from the 
% pain of mastering matrix operations while preparing "myfunction.m".

figure('NumberTitle', 'off', 'Name', VersionControl, 'Units', 'normalized', ...
    'Position', [(1 - WindowPositionScale) / 2, ...
    (1 - WindowPositionScale) / 2, WindowPositionScale, WindowPositionScale]);
% Give the figure an eye-friendly size regardless of screen resolutions.
% 'OuterPosition' is better than 'Position', but does not work on Octave.
% It is good practice to restore the default unit of the figure, which is  
% fine on Matlab; but on Octave after the unit is restored to its default
% value, the figure changes back to its original size.

if isLogarithmicX
    semilogx(x, y, 'k', 'LineWidth', 1.5, 'Tag', 'FunctionCurve'); 
else
    plot(x, y, 'k', 'LineWidth', 1.5, 'Tag', 'FunctionCurve');
end

set(gca, 'FontSize', 12, 'NextPlot', 'add', 'XGrid', 'on', 'YGrid', 'on', 'Box', 'on');
set(gca, 'UserData', [x; y]);
XLimit = get(gca, 'XLim');

plot(XLimit, [0 0], 'r', 'LineWidth', 2);
% Draw a zero-Y line for assistance.

if isOctave
    % do nothing.
else
    temp = get(gca, 'Children'); set(gca, 'Children', temp(end:-1:1));
end
% Make the curve of the funtion on top of the zero-Y line by changing the
% order of the children of the axes.  But such change causes the title to
% disappear on Octave.

nZeroCrossings = sum(abs(diff((y > 0) + (y <= 0) * -1))) / 2;
% The number of zero-crossings of the function.
if nZeroCrossings == 1
    set(gca, 'Color', PermissionColor);
else
    set(gca, 'Color', ProhibitionColor);
end

Heading = {['{\color{red}Instructions}: By zooming or panning, ' ...
    'find an interval in which the curve in black traverses only once ' ...
    'the horizontal line in red, and the axis turns green ']; ['accordingly, ' ...
    'then close the figure to start the iteration.  If such a condition is ' ...
    'not satisfied, closing the figure does not start the iteration, and ' ...
    'the axis is pink.']};

set(gcf, 'UserData', Heading);

title(Heading, 'FontSize', 12); 

if isOctave
    addlistener(gca, 'XLim', @showhints);
    addlistener(gca, 'YLim', @showhints);
else
    try
        h1 = zoom; 
        h2 = pan;
        set(h1, 'ActionPreCallback', @hidetitle, 'ActionPostCallback', @showhints);
        set(h2, 'ActionPreCallback', @hidetitle, 'ActionPostCallback', @showhints);
    end
    % Some version of Matlab does not support post actions of zooming or
    % panning.
end

set(gcf, 'CloseRequestFcn', @iteration);


% Zooming or panning changes the axes limits, and then the axis updates 
% the position of its title.  For a complex title, the update is a bit 
% time-consuming, which causes the zooming or panning sluggish.  In order
% to avoid this, the following function replaces the complex title with a 
% simple one before zooming or panning is performed.  When the zooming or 
% panning is finished, the original (complex) title is restored.
function hidetitle(obj, evd)
title({' '; ' '}, 'FontSize', 12); 


function showhints(varargin)
% this function works for both Matlab and Octave despite the fact that 
% Matlab has two arguments (obj and evd) and Octave has one (obj)

global PermissionColor ProhibitionColor

title(get(gcf, 'UserData'), 'FontSize', 12);

CurveData = get(gca, 'UserData');
XLimit = get(gca, 'XLim');

LeftBorder = find((CurveData(1, :) > XLimit(1)), 1, 'first');
RightBorder = find((CurveData(1, :) < XLimit(2)), 1, 'last');

temp = (CurveData(2, LeftBorder:RightBorder) > 0) + ...
    (CurveData(2, LeftBorder:RightBorder) <= 0) * -1;
nZeroCrossings = sum(abs(diff(temp))) / 2;
% The number of zero-crossings of the function.
if nZeroCrossings == 1
    set(gca, 'Color', PermissionColor);
else
    set(gca, 'Color', ProhibitionColor);
end

if (RightBorder - LeftBorder) <= 10
    set(findobj('Tag', 'FunctionCurve'), 'Marker', 'O', 'LineStyle', '--');
else
    set(findobj('Tag', 'FunctionCurve'), 'Marker', 'none', 'LineStyle', '-');
end

if LeftBorder == RightBorder
    warndlg('There is only one point in this interval.  Try zooming out.', 'Zooming-in Warning', 'modal');
end


% Call Matlab native function "fzero" to find roots of the function.
function iteration(src, evnt)
global ProhibitionColor isNewVersionAvailable Webpage isOctave

SmallestExponentAllowed = -14;
SmallestExponentPreferred = -11;
% The values of the two variables are set according to my experiences.  
% According to my experences, if the root found by fzero has a smaller 
% exponent than -14, the accuracy is unacceptable; if it has a greater 
% exponent than -11, the accuracy is fine.

set(gcf, 'CloseRequestFcn', 'delete(gcf)');

if get(gca, 'Color') == ProhibitionColor
    disp(' ');
    disp('Error: this interval encompasses no or more than one root.');
    disp('       bisection cannot be performed in those cases.');
    disp(' ');
    disp('Read the instruction and try again.');
else
    CurveData = get(gca, 'UserData');
    XLimit = get(gca, 'XLim');
    
    LeftBorder = find((CurveData(1, :) > XLimit(1)), 1, 'first');
    RightBorder = find((CurveData(1, :) < XLimit(2)), 1, 'last');
    
    try
        [Root, Evaluation, ExitFlag] = fzero(@myfunction, ...
            CurveData(1, [LeftBorder, RightBorder]));
        if ExitFlag == 1
            Exponent = floor(log10(abs(Root)));
            if Exponent < SmallestExponentAllowed
                shift = Exponent - SmallestExponentPreferred;
                disp(' ');
                disp('!!!WARNING!!!')
                disp(' ');
                disp(['The root (' num2str(Root) ') is significantly small, ' ...
                    'so it might be inaccurate.']);
                disp(['Add the following line to the beginning of the ' ...
                    'function codes, and solve it again.']);
                
                AddedLine = ['x = x * 1e' num2str(shift) ';'];
                if ~isOctave
                    disp('(this line has been copied to the clipboard)');
                    clipboard('copy', AddedLine);
                end
                
                disp(' ');
                disp(AddedLine);
                disp(' ');
                disp(['After you have the new root, subtract ' ...
                    num2str(-shift) ' from its exponent.']);
            else
                disp(['     Root: ' num2str(Root, '%17.16e')]);
                disp(['     Function evaluated at the root: ' num2str(Evaluation)]);
                if Root > 0
                    disp(['     The logarithm of the root: ' num2str(log10(Root))]);
                end
            end
        else
            disp('     The iterative process failed to find a root.')
        end
    catch ME
        disp(' ');
        disp(['Error: ' ME.message]);
        disp(' ');
        disp('Read the instruction and try again.');
    end
end

delete(gcf);

if isNewVersionAvailable
    pause(3);
    
    if isOctave
        disp(['A new version of iroots was found.  ' ...
            'Please go to the following webpage to download.']);
        disp(Webpage);
    else
        Message = ['A new version of iroots was found.\n' ...
            'Do you want to visit the website to download it?'];
        Answer = questdlg(sprintf(Message), 'Version', 'Yes', 'No', 'Yes');
        
        if strcmp(Answer, 'Yes')
            web(Webpage, '-browser');
        end
    end
end

evalin('base', 'clear global PermissionColor Webpage isOctave ProhibitionColor isNewVersionAvailable');


% Based on a user-input value, try to find a valid interval for the
% bisection method.
function Interval = bingo(a)
Interval = [];
MaxIterations = 1000;

for i = 1: MaxIterations
    isLongLoop = false;
    
    tic;
    while isinf(myfunction(a))
        a = a + eps;
        
        ElapsedTime = toc;
        if ElapsedTime > 5
            isLongLoop = true;
            break;
        end
    end
    if isLongLoop
        break;
    end
    % search a finite value of the function.  If it takes more than 5 sec,
    % the search may be too long.
    
    f = myfunction(a);
    
    Delta = eps;
    tic;
    while myfunction(a + Delta) == f
        Delta = Delta * 10;
        
        ElapsedTime = toc;
        if ElapsedTime > 5
            isLongLoop = true;
            break;
        end
    end
    if isLongLoop
        break;
    end
    % search a small step that effects the calculation of first derivative.
    % If it takes more than 5 sec, the search may be too long.
    
    Step = f * Delta / (myfunction(a + Delta) - f)*1.01;
    b = a - Step;
    % a new value with respect to a of the independent variable that alters 
    % the function in the direction of sign change.
    % 1.01 is an arbitrary value to avoid the trap of Newton method.
    
    tic;
    while isinf(myfunction(b))
        b = b + eps;
        
        ElapsedTime = toc;
        if ElapsedTime > 5
            isLongLoop = true;
            break;
        end
    end
    if isLongLoop
        break;
    end
    % search a finite value of the function.  If it takes more than 5 sec,
    % the search may be too long.
    
    if f * myfunction(b) < 0
        % bingo!
        LeftBorder = min(a, b);
        k = 10^sign(LeftBorder);
        LeftBorder = LeftBorder / k;
        RightBorder = max(a, b);
        k = 10^sign(RightBorder);
        RightBorder = RightBorder * k;
        Interval = [LeftBorder RightBorder];
        % purposely enlarge the interval to make it less horrifying.
        break;
    else
        a = b;
    end
end


function Templates = ProcessTemplates(TemplateFileName)
if isequal(exist(TemplateFileName, 'file'), 2)
    FileID = fopen(TemplateFileName);
    Contents = textscan(FileID, '%s', 'delimiter', '\n');
    Contents = Contents{:};
    fclose(FileID);
    
    nLines = length(Contents);
    aux = zeros(nLines, 1);
    for i = 1:nLines-1
        aux(i) = (~isempty(Contents{i})) && (Contents{i}(1) == '#');
    end
    Positions = find(aux);
    % Find all the lines that start with "#" except the last one.  The 
    % last line is skipped for the convenience of subsequent programing.
    
    if isempty(Positions) || any(diff(Positions) == 1)
        disp(['An error is found in the template file "' TemplateFileName '".']);
        fprintf(['A template must start with a separate line that must start ' ...
            'with "#", \nand there must be only one such line for a template.\n']);
        Templates = [];
    else
        Templates.Positions = Positions;
        Templates.Contents = Contents;
        % templates are successfully loaded.
    end
else
    Templates = []; % The template file does not exist.
    
    disp('You can create templates for common functions.');
    disp('When you solve such an equation afterwards, just load');
    disp('corresponding template and input some values.');
    disp(' ');
    disp(['Templates are stored in a text file named "' TemplateFileName '".']);
    disp('An example file was just created for your reference.  When you');
    disp('open the file, you will find it is fairly self-explanatory, just');
    disp('keep in mind that every template must start with a separate line');
    disp('that must start with "#".');
    
    FileID = fopen(TemplateFileName, 'w');
    fprintf(FileID, '%s\n', '#[H+] of a monoprotic acid');
    fprintf(FileID, '%s\n', 'k = ;');
    fprintf(FileID, '%s\n', 'c = ;');
    fprintf(FileID, '%s\n', 'y = x - k/(x + k)*c - 1e-14/x;');
    fprintf(FileID, '%s\n', '');
    fprintf(FileID, '%s\n', '#[H+] of a diprotic acid');
    fprintf(FileID, '%s\n', 'k1 = ; k2 = ;');
    fprintf(FileID, '%s\n', 'c = ;');
    fprintf(FileID, '%s\n', 'y = x - (k1*x + 2*k1*k2)/(x^2 + k1*x + k1*k2)*c - 1e-14/x;');
    fprintf(FileID, '%s\n', '');
    fprintf(FileID, '%s\n', '#[H+] of a triprotic acid');
    fprintf(FileID, '%s\n', 'k1 = ; k2 = ; k3 = ;');
    fprintf(FileID, '%s\n', 'c = ;');
    fprintf(FileID, '%s\n', 'y = x - (k1*x^2 + 2*k1*k2*x + 3*k1*k2*k3)/(x^3 + k1*x^2 + k1*k2*x + k1*k2*k3)*c - 1e-14/x;');
    fclose(FileID);
end